"""
Written by Matteo Dunnhofer - 2020

Class that defines the TRAS, TRAST, TRASFUST trackers
"""
#import sys
#sys.path.insert(0, '..')
import os
import copy
import numpy as np
import torch
from config import Configuration
from model.ACModel import ActorCriticModel
from data.DataTransformer import DataTransformer
import utils as ut
from demotrackers.SiamFC.siamfc import TrackerSiamFC
#from demotrackers.pytracking2.pytracking.pytracking.TrackerATOM import TrackerATOM
#from demotrackers.pytracking2.pytracking.pytracking.TrackerDiMP import TrackerDiMP
#from demotrackers.DemoTracker import ResultsTracker

class TRAS(object):

    def __init__(self, ckpt_path, lstm_update=False, gpu_id=0, save_img_preds=False):
        super(TRAS, self).__init__()

        self.cfg = Configuration()
        self.data_transformer = DataTransformer(self.cfg)

        self.save_img_preds = save_img_preds

        self.gpu_id = gpu_id
        self.device = torch.device('cuda', self.gpu_id) if self.cfg.USE_GPU else torch.device('cpu')

        self.model = ActorCriticModel(self.cfg).to(self.device)
        self.lstm_update = lstm_update

        ckpt = torch.load(ckpt_path, map_location=self.device)
        if type(ckpt) is dict:
            ckpt = ckpt['model_state_dict']
        self.model.load_state_dict(ckpt)

        self.model.eval()

    def init(self, image, box, **kwargs):
        image = image.convert('RGB')

        self.curr_bb = box #copy.deepcopy(box)
        self.prev_img = image #copy.deepcopy(image)

        self.model_state = self.model.init_state(self.device) #copy.deepcopy(self.model.init_state(self.device))

        self.step = 1

    def update(self, image):
        image = image.convert('RGB')

        ## LSTM STATE UPDATE
        if self.lstm_update and (self.step % self.cfg.SEQ_LENGTH == 0):
            #self.model_state = copy.deepcopy(self.model.init_state(self.device))
            self.model_state = self.first_model_state

        bb = ut.get_crop_bb(copy.deepcopy(self.curr_bb), image.size[0], image.size[1], 1.5)

        state1 = self.data_transformer.preprocess_img(self.prev_img, bb).to(self.device).unsqueeze(0)
        state2 = self.data_transformer.preprocess_img(image, bb).to(self.device).unsqueeze(0)

        policy_mu, _, self.model_state = self.model(state1, state2, self.model_state, self.device)
        mu = torch.clamp(policy_mu, -1.0, 1.0)

        action = ut.denorm_action(mu.data.cpu().numpy()[0], self.curr_bb)

        if self.step == 1:
            self.first_model_state = (self.model_state[0].clone(), self.model_state[1].clone())

        self.step += 1

        self.curr_bb = ut.clip_bb(action, image) #copy.deepcopy(action)
        self.prev_img = image #copy.deepcopy(image)

        return np.array(self.curr_bb)

class TRAST(object):

    def __init__(self, ckpt_path, teacher_ids, return_ctrl_idx=False, lstm_update=False, use_real_trackers=True, save_img_preds=False, gpu_id=0, eval_decisions=False, teacher_training='POOL', dset_name=''):
        super(TRAST, self).__init__()

        self.cfg = Configuration()
        self.data_transformer = DataTransformer(self.cfg)

        self.save_img_preds = save_img_preds
        self.return_ctrl_idx = return_ctrl_idx

        self.gpu_id = gpu_id
        self.device = torch.device('cuda', self.gpu_id) if self.cfg.USE_GPU else torch.device('cpu')

        self.model = ActorCriticModel(self.cfg).to(self.device)
        self.lstm_update = lstm_update

        ckpt = torch.load(ckpt_path, map_location=self.device)
        if type(ckpt) is dict:
            ckpt = ckpt['model_state_dict']
        self.model.load_state_dict(ckpt)
        self.model.eval()

        self.teacher_ids = teacher_ids
        self.teacher_trackers = []
        for teacher_id in teacher_ids:

            if use_real_trackers:
                if teacher_id == 'SiamFC':
                    teacher_tracker = TrackerSiamFC(gpu_id=self.gpu_id)
                elif teacher_id == 'ECO':
                    from demotrackers.pytracking2.pytracking.pytracking.TrackerECO import TrackerECO
                    teacher_tracker = TrackerECO(gpu_id=self.gpu_id)
                elif teacher_id == 'DiMP':
                    from demotrackers.pytracking2.pytracking.pytracking.TrackerDiMP import TrackerDiMP
                    teacher_tracker = TrackerDiMP(gpu_id=self.gpu_id)
                elif teacher_id == 'ATOM':
                    from demotrackers.pytracking2.pytracking.pytracking.TrackerATOM import TrackerATOM
                    teacher_tracker = TrackerATOM(gpu_id=self.gpu_id)

            self.teacher_trackers.append(teacher_tracker)

        self.n_teachers = len(self.teacher_trackers)

        self.eval_decisions = eval_decisions
        if self.eval_decisions:
            self.n_decisions = np.zeros(len(teacher_ids)+1)
            self.tot_n_decisions = 0
            self.f = open('TRAST-t' + teacher_training + '-e' + str(teacher_ids) + '-decisions-' + dset_name + '.txt', 'w')

    def init(self, image, box, **kwargs):
        # initialize teacher trackers
        self.teachers_failed_init = [False] * self.n_teachers
        for i, et in enumerate(self.teacher_trackers):
            try:
                et.init(image, box, **kwargs)
            except Exception as e:
                print(e)
                print('Teacher {} failure on initialization'.format(et.name))
                self.teachers_failed_init[i] = True

        image = image.convert('RGB')

        self.curr_bb = copy.deepcopy(box)

        self.prev_img = copy.deepcopy(image)

        # init TRAST lstm states and previous bounding boxes
        self.model_state = copy.deepcopy(self.model.init_state(self.device))
        self.teacher_model_states = []
        self.teacher_curr_bbs = []
        for _ in range(len(self.teacher_trackers)):
            self.teacher_model_states.append(copy.deepcopy(self.model.init_state(self.device)))
            self.teacher_curr_bbs.append(copy.deepcopy(box))

        self.step = 1


    def update(self, image):
        # get teachers predictions
        teacher_preds = np.zeros((self.n_teachers, 4), dtype=np.float32)
        teachers_failed = [False] * self.n_teachers

        for i, et in enumerate(self.teacher_trackers):
            try:
                teacher_pred = et.update(image)
                teacher_preds[i] = np.array(teacher_pred)
            except Exception as e:
                print('teacher {} failure'.format(et.name))
                teachers_failed[i] = True
                teacher_preds[i] = np.zeros(4)

        image = image.convert('RGB')

        # update lstm state
        if self.lstm_update and (self.step % self.cfg.SEQ_LENGTH == 0):
            self.model_state = self.first_model_state

            for i in range(self.n_teachers):
                if (not self.teachers_failed_init[i]):
                    self.teacher_model_states[i] = (self.first_model_states[i][0].clone(), self.first_model_states[i][1].clone())

        # student prediction
        bb = ut.get_crop_bb(copy.deepcopy(self.curr_bb), image.size[0], image.size[1], 1.5)

        state1 = self.data_transformer.preprocess_img(self.prev_img, bb).to(self.device).unsqueeze(0)
        state2 = self.data_transformer.preprocess_img(image, bb).to(self.device).unsqueeze(0)

        # get student tracker bounding box estimate
        policy_mu, value, self.model_state = self.model(state1, state2, self.model_state, self.device)
        mu = torch.clamp(policy_mu, -1.0, 1.0)
        action = ut.denorm_action(mu.data.cpu().numpy()[0], self.curr_bb)

        # evaluate each tracker prediction
        teacher_values = np.zeros(self.n_teachers) - 10000
        for i in range(self.n_teachers):
            if (not self.teachers_failed_init[i]):
                bb = ut.get_crop_bb(copy.deepcopy(self.teacher_curr_bbs[i]), image.size[0], image.size[1], 1.5)

                prev_img = copy.deepcopy(self.prev_img)
                curr_img = copy.deepcopy(image)

                state1 = self.data_transformer.preprocess_img(prev_img, bb).to(self.device).unsqueeze(0)
                state2 = self.data_transformer.preprocess_img(curr_img, bb).to(self.device).unsqueeze(0)

                # get student value
                _, teacher_value, n_teacher_model_state = self.model(state1, state2, self.teacher_model_states[i], self.device)
                self.teacher_model_states[i] = (n_teacher_model_state[0].clone(), n_teacher_model_state[1].clone())
                teacher_values[i] = teacher_value.data.cpu().numpy()[0, 0]

                self.teacher_curr_bbs[i] = copy.deepcopy(teacher_preds[i])


        self.prev_img = copy.deepcopy(image)

        # choose the teacher output with the highest value
        max_teacher_idx = np.argmax(teacher_values)

        if (not self.teachers_failed_init[max_teacher_idx]) and (not teachers_failed[max_teacher_idx]):
            # evaluate the quality of the TRASracker and the teacher and chhose one of the two
            if value.data.cpu().numpy()[0,0] >= teacher_values[max_teacher_idx]:
                self.curr_bb = copy.deepcopy(action)
                ctrl_idx = 0
            else:
                self.curr_bb = copy.deepcopy(teacher_preds[max_teacher_idx])
                ctrl_idx = 1
        else:
            # handling teacher failure
            self.curr_bb = copy.deepcopy(action)
            ctrl_idx = 0

        self.curr_bb = ut.clip_bb(self.curr_bb, image)

        if self.step == 1:
            self.first_model_state = (self.model_state[0].clone(), self.model_state[1].clone())

            self.first_model_states = []
            for i in range(self.n_teachers):
                self.first_model_states.append((self.teacher_model_states[i][0].clone(), self.teacher_model_states[i][1].clone()))

        self.step += 1

        #print(self.step, self.curr_bb)
        if self.eval_decisions:
            if ctrl_idx > 0:
                self.n_decisions[max_teacher_idx] += 1
            else:
                self.n_decisions[-1] += 1

            self.tot_n_decisions += 1

            self.f.seek(0)

            self.f.write(''.join(['{} n: {} ({}%)'.format(en, self.n_decisions[i], (self.n_decisions[i] / self.tot_n_decisions) * 100) for i, en in enumerate(self.teacher_ids + ['TRAS'])])
                        + ' - total n {}'.format(self.tot_n_decisions))

            self.f.flush()

        if self.return_ctrl_idx:
            return self.curr_bb, ctrl_idx
        else:
            return self.curr_bb

class TRASFUST(object):

    def __init__(self, ckpt_path, teacher_ids, return_ctrl_idx=False, lstm_update=False, use_real_trackers=True, save_img_preds=False, gpu_id=0, eval_decisions=False, teacher_training='POOL', dset_name=''):
        super(TRASFUST, self).__init__()

        self.cfg = Configuration()
        self.data_transformer = DataTransformer(self.cfg)

        self.save_img_preds = save_img_preds
        self.return_ctrl_idx = return_ctrl_idx

        self.gpu_id = gpu_id
        self.device = torch.device('cuda', self.gpu_id) if self.cfg.USE_GPU else torch.device('cpu')

        self.model = ActorCriticModel(self.cfg).to(self.device)
        self.lstm_update = lstm_update

        ckpt = torch.load(ckpt_path, map_location=self.device)
        if type(ckpt) is dict:
            ckpt = ckpt['model_state_dict']
        self.model.load_state_dict(ckpt)
        self.model.eval()

        self.teacher_ids = teacher_ids
        self.teacher_trackers = []
        for teacher_id in teacher_ids:

            if use_real_trackers:
                if teacher_id == 'SiamFC':
                    teacher_tracker = TrackerSiamFC(gpu_id=self.gpu_id)
                elif teacher_id == 'ECO':
                    from demotrackers.pytracking2.pytracking.pytracking.TrackerECO import TrackerECO
                    teacher_tracker = TrackerECO(gpu_id=self.gpu_id)
                elif teacher_id == 'DiMP':
                    from demotrackers.pytracking2.pytracking.pytracking.TrackerDiMP import TrackerDiMP
                    teacher_tracker = TrackerDiMP(gpu_id=self.gpu_id)
                elif teacher_id == 'ATOM':
                    from demotrackers.pytracking2.pytracking.pytracking.TrackerATOM import TrackerATOM
                    teacher_tracker = TrackerATOM(gpu_id=self.gpu_id)

            self.teacher_trackers.append(teacher_tracker)

        self.n_teachers = len(self.teacher_trackers)

        self.eval_decisions = eval_decisions
        if self.eval_decisions:
            self.n_decisions = np.zeros(len(teacher_ids))
            self.tot_n_decisions = 0
            self.f = open('TRASFUST-t' + teacher_training + '-e' + str(teacher_ids) + '-decisions-' + dset_name + '.txt', 'w')

    def init(self, image, box, **kwargs):
        # initialize teacher trackers
        self.teachers_failed_init = [False] * self.n_teachers
        for i, et in enumerate(self.teacher_trackers):
            try:
                et.init(image, box, **kwargs)
            except Exception as e:
                print(e)
                print('teacher {} failure on initialization'.format(et.name))
                self.teachers_failed_init[i] = True

        image = image.convert('RGB')

        self.curr_bb = copy.deepcopy(box)

        self.prev_img = copy.deepcopy(image)

        # init TRASFUST lstm states and previous bounding boxes
        self.teacher_model_states = []
        self.teacher_curr_bbs = []
        for _ in range(len(self.teacher_trackers)):
            self.teacher_model_states.append(copy.deepcopy(self.model.init_state(self.device)))
            self.teacher_curr_bbs.append(copy.deepcopy(box))

        self.step = 1


    def update(self, image):
        # get teachers predictions
        teacher_preds = np.zeros((self.n_teachers, 4), dtype=np.float32)
        teachers_failed = [False] * self.n_teachers

        for i, et in enumerate(self.teacher_trackers):
            try:
                teacher_pred = et.update(image)
                teacher_preds[i] = np.array(teacher_pred)
            except Exception as e:
                print('teacher {} failure'.format(et.name))
                teachers_failed[i] = True
                teacher_preds[i] = np.zeros(4)


        image = image.convert('RGB')

        # update lstm state
        if self.lstm_update and (self.step % self.cfg.SEQ_LENGTH == 0):
            for i in range(self.n_teachers):
                if (not self.teachers_failed_init[i]):
                    self.teacher_model_states[i] = (self.first_model_states[i][0].clone(), self.first_model_states[i][1].clone())

        # evaluate each tracker prediction with the student
        teacher_values = np.zeros(self.n_teachers) - 10000
        for i in range(self.n_teachers):
            if (not self.teachers_failed_init[i]):
                bb = ut.get_crop_bb(copy.deepcopy(self.teacher_curr_bbs[i]), image.size[0], image.size[1], 1.5)

                prev_img = copy.deepcopy(self.prev_img)
                curr_img = copy.deepcopy(image)

                state1 = self.data_transformer.preprocess_img(prev_img, bb).to(self.device).unsqueeze(0)
                state2 = self.data_transformer.preprocess_img(curr_img, bb).to(self.device).unsqueeze(0)

                # get student value
                _, teacher_value, n_teacher_model_state = self.model(state1, state2, self.teacher_model_states[i], self.device)
                self.teacher_model_states[i] = (n_teacher_model_state[0].clone(), n_teacher_model_state[1].clone())
                teacher_values[i] = teacher_value.data.cpu().numpy()[0, 0]

                self.teacher_curr_bbs[i] = copy.deepcopy(teacher_preds[i])

        self.prev_img = copy.deepcopy(image)

        # choose the teacher output with the highest value
        max_teacher_idx = np.argmax(teacher_values)
        self.curr_bb = copy.deepcopy(teacher_preds[max_teacher_idx])
        self.curr_bb = ut.clip_bb(self.curr_bb, image)

        if self.step == 1:
            self.first_model_states = []
            for i in range(self.n_teachers):
                self.first_model_states.append((self.teacher_model_states[i][0].clone(), self.teacher_model_states[i][1].clone()))

        self.step += 1

        if self.eval_decisions:
            self.n_decisions[max_teacher_idx] += 1
            self.tot_n_decisions += 1

            self.f.seek(0)

            self.f.write(''.join(['{} n: {} ({}%)'.format(en, self.n_decisions[i], (self.n_decisions[i] / self.tot_n_decisions) * 100) for i, en in enumerate(self.teacher_ids)])
                        + ' - total n {}'.format(self.tot_n_decisions))

            self.f.flush()

        if self.return_ctrl_idx:
            return self.curr_bb, max_teacher_idx
        else:
            return self.curr_bb
