"""
Written by Matteo Dunnhofer - 2020

Utility functions
"""
import numpy as np

def clip(value, min_value, max_value):
    return max(min(value, max_value), min_value)

def cxcywh2xyxy(bbox):
    x1 = bbox[0] - (bbox[2] / 2)
    y1 = bbox[1] - (bbox[3] / 2)

    return [x1, y1, x1 + bbox[2], y1 + bbox[3]]

def xywh2xyxy(bbox):
    x1 = bbox[0]
    y1 = bbox[1]

    return [x1, y1, x1 + bbox[2], y1 + bbox[3]]

def xyxy2cxcywh(bbox):
    w = bbox[2] - bbox[0]
    h = bbox[3] - bbox[1]
    x = bbox[0] + (w / 2)
    y = bbox[1] + (h / 2)

    return [x, y, w, h]

def xyxy2xywh(bbox):
    w = bbox[2] - bbox[0]
    h = bbox[3] - bbox[1]
    x = bbox[0]
    y = bbox[1]

    return [x, y, w, h]


def clip_bb(bbox, image):
    x = clip(bbox[0], 0, image.size[0])
    y = clip(bbox[1], 0, image.size[1])
    w = clip(bbox[2], 0, image.size[0])
    h = clip(bbox[3], 0, image.size[1])

    bbox = np.array([x, y, w, h])

    return bbox

def get_crop_bb(bb, img_w, img_h, k):
    # compute the coordinates of a padded crop

    w = clip(bb[2], 0, img_w)
    h = clip(bb[3], 0, img_h)

    cx = clip(bb[0] + (w / 2), 0, img_w)
    cy = clip(bb[1] + (h / 2), 0, img_h)

    new_w = k * w
    new_h = k * h

    new_x1 = cx - (new_w / 2)
    new_y1 = cy - (new_h / 2)
    new_x2 = cx + (new_w / 2)
    new_y2 = cy + (new_h / 2)

    crop = [new_x1, new_y1, new_w, new_h]

    return crop

def denorm_action(action, old_bb):
    # from shift to bounding box coordinates
    norm_coeff = 1.0
    new_x = old_bb[0] + (norm_coeff * action[0] * old_bb[2])
    new_y = old_bb[1] + (norm_coeff * action[1] * old_bb[3])
    new_w = old_bb[2] + (norm_coeff * action[2] * old_bb[2])
    new_h = old_bb[3] + (norm_coeff * action[3] * old_bb[3])

    return [new_x, new_y, new_w, new_h]

def make_full_size(x, output_sz):
    '''
    zero-pad input x (right and down) to match output_sz
    x: numpy array e.g., binary mask
    output_sz: size of the output [width, height]
    '''
    if x.shape[0] == output_sz[1] and x.shape[1] == output_sz[0]:
        return x
    pad_x = output_sz[0] - x.shape[1]
    if pad_x < 0:
        x = x[:, :x.shape[1] + pad_x]
        # padding has to be set to zero, otherwise pad function fails
        pad_x = 0
    pad_y = output_sz[1] - x.shape[0]
    if pad_y < 0:
        x = x[:x.shape[0] + pad_y, :]
        # padding has to be set to zero, otherwise pad function fails
        pad_y = 0
    return np.pad(x, ((0, pad_y), (0, pad_x)), 'constant', constant_values=0)

def _rect_from_mask(mask):
    x_ = np.sum(mask, axis=0)
    y_ = np.sum(mask, axis=1)
    x0 = np.min(np.nonzero(x_))
    x1 = np.max(np.nonzero(x_))
    y0 = np.min(np.nonzero(y_))
    y1 = np.max(np.nonzero(y_))
    return [x0, y0, x1 - x0 + 1, y1 - y0 + 1]